@@ -8,7 +8,7 @@
 
 #include <errno.h>
 #include <stdio.h>
-#include <stdlib.h>
+#include <fcntl.h>
 #include <string.h>
 #include <unistd.h>
 #include <sys/dir.h>
@@ -18,6 +18,7 @@
 #include "xchpst.h"
 #include "options.h"
 #include "mount.h"
+#include "compat/mount_setattr.h"
 
 int special_mount(char *path, char *fs, char *desc, char *options) {
   const char *op = "mkdirat";
