@@ -19,6 +19,9 @@
 #include "options.h"
 
 struct options opt;
+struct runtime runtime = { 0 };
+
+static const char HIDE = (char) -1;
 
 const struct option_info options_info[] = {
   { C_R, OPT_SETUIDGID,   'u',  NULL,       required_argument, "set uid, gid and supplementary groups", "[:]USER[:GROUP]*", },
@@ -55,17 +58,22 @@ const struct option_info options_info[] = {
   { C_ALL,OPT_VERBOSE,    'v',  "verbose",  no_argument,       "be verbose", NULL },
   { C_R, OPT_VERSION,     'V',  "version",  no_argument,       "show " NAME_STR " version", NULL },
   { C_R, OPT_PGRPHACK,    'P',  NULL,       no_argument,       "run in new process group", NULL },
+  { C_R, OPT_FGHACK,      '\0', "fghack",   no_argument,       "wait for all spawned processes to terminate", NULL },
   { C_R, OPT_CLOSE_STDIN, '0',  NULL,       no_argument,       "close stdin", NULL },
   { C_R, OPT_CLOSE_STDOUT,'1',  NULL,       no_argument,       "close stdout", NULL },
   { C_R, OPT_CLOSE_STDERR,'2',  NULL,       no_argument,       "close stderr", NULL },
   { C_X, OPT_LEGACY,      '@',  NULL,       no_argument,       "restricts following options to chpst(8) ones", NULL },
+  { C_X, OPT_OPT_START,   HIDE, "opt[",     no_argument,       "starts a group of optional options", NULL },
+  { C_X, OPT_OPT_END,     HIDE, "opt]",     no_argument,       "ends a group of optional options", NULL },
   { C_X, OPT_HELP,        'h',  "help",     no_argument,       "show help", NULL },
   { C_X, OPT_FILE,        '\0', "file",     required_argument, "read options from file", "FILE" },
   { C_X, OPT_EXIT,        '\0', "exit",     optional_argument, "exit (with optional RETCODE)", "RETCODE" },
   { C_X, OPT_MOUNT_NS,    '\0', "mount-ns", no_argument,       "create mount namespace", NULL },
   { C_X, OPT_NET_NS,      '\0', "net-ns",   no_argument,       "create net namespace", NULL },
   { C_X, OPT_USER_NS,     '\0', "user-ns",  no_argument,       "create user namespace", NULL },
-  { C_X, OPT_PID_NS,      '\0', "pid-ns",   no_argument,       "create pid namespace", NULL },
+  { COMPAT_FGHACK,
+         OPT_PID_NS,      'I',  "pid-ns",   no_argument,       "create pid namespace", NULL },
+  { C_R, OPT_PID_NS,      '\0', "pid-ns",   no_argument,       "create pid namespace", NULL },
   { C_X, OPT_UTS_NS,      '\0', "uts-ns",   no_argument,       "create uts namespace", NULL },
   { C_X, OPT_NET_ADOPT,   '\0', "adopt-net",required_argument, "adopt net namespace", "NS-PATH" },
   { C_X, OPT_PRIVATE_RUN, '\0', "private-run",  no_argument,    "create private /run", NULL },
@@ -95,6 +103,7 @@ const struct option_info options_info[] = {
   { C_X, OPT_LOGIN,       '\0', "login",     no_argument,      "simulate login environment", NULL },
   { C_X, OPT_OOM,         '\0', "oom",       required_argument,"set oom adjust value", "ADJ" },
   { C_X, OPT_HARDLIMIT,   '\0', "hardlimit", no_argument,      "set hard limits with soft limits", NULL },
+  { C_X, OPT_DEV_NULL,    '\0', "dev-null",  no_argument,      "open /dev/null instead of closing fds", NULL },
   { C_A, OPT_UIDNUM,      'u',  NULL,        required_argument,"set numerical uid", "UID" },
   { C_A, OPT_GIDNUM,      'g',  NULL,        required_argument,"set numerical gid", "GID" },
   { C_A, OPT_GIDSNUM,     'G',  NULL,        required_argument,"set numerical supplementary groups", "GID*" },
@@ -102,6 +111,7 @@ const struct option_info options_info[] = {
   { C_X, OPT_UGFROMENV,   '\0', "ugids-from-env",  no_argument,"set numerical uid, gid and supplementary groups from env", NULL },
   { C_A, OPT_UGCLEARENV,  'z',  NULL,              no_argument,"clear numerical uid, gid and supplementary groups in env", NULL },
   { C_X, OPT_UGCLEARENV,  '\0', "ugids-clear-env", no_argument,"clear numerical uid, gid and supplementary groups in env", NULL },
+  { C_X, OPT_IDMAP,       '\0', "id-map",    required_argument,"define uid/gid mapping to use for user namespace", "self|identity|full" },
 };
 #define max_options ((ssize_t) ((sizeof options_info / sizeof *options_info)))
 
@@ -117,7 +127,7 @@ void options_print(FILE *out) {
   int i = 0;
 
   for (optdef = options_info; optdef - options_info < max_options; optdef++) {
-    if (optdef->compat_level & opt.app->compat_level) {
+    if (is_compatible(optdef) && optdef->short_name != HIDE) {
       if (i++ == 0)
         fprintf(out, "\n OPTIONS\n");
       fprintf(out, "  %c%c%s%s%s%c%-*s %s\n",
@@ -177,7 +187,7 @@ bool options_init(void) {
   int i = 0;
 
   for (optdef = options_info; optdef - options_info < max_options; optdef++) {
-    if (optdef->compat_level & opt.app->compat_level) {
+    if (is_compatible(optdef) && optdef->short_name != HIDE) {
 
       /* Populate long option list */
       if (optdef->long_name) {
@@ -214,7 +224,7 @@ bool options_init(void) {
   optstr_len = 0;
   optstr[optstr_len++] = '+'; /* Stop accepting options after prog name */
   for (optdef = options_info; optdef - options_info < max_options; optdef++) {
-    if (optdef->compat_level & opt.app->compat_level && optdef->short_name) {
+    if (is_compatible(optdef) && optdef->short_name != 0 && optdef->short_name != HIDE) {
       optstr[optstr_len++] = optdef->short_name;
       switch (optdef->has_arg) {
       case optional_argument:
@@ -311,8 +321,13 @@ bool parse_caps(cap_bits_t *caps, char *names) {
       good = false;
     }
     set |= (1ull << val);
+    if (is_debug())
+      fprintf(stderr, "  added capability %d(%s) to a set\n", val, cap_to_name(val));
   }
   *caps = set;
+  if (is_debug())
+    fprintf(stderr, "  set is: %llx\n", (long long) *caps);
+
   return good;
 }
 #endif
@@ -366,8 +381,8 @@ void parse_cpus(char *spec) {
       goto fail;
     end = tok;
   }
-  opt.cpu_affinity.size = CPU_ALLOC_SIZE(max++);
-  opt.cpu_affinity.mask = CPU_ALLOC(max);
+  opt.cpu_affinity.size = CPU_ALLOC_SIZE(max);
+  opt.cpu_affinity.mask = CPU_ALLOC(max + 1);
   if (opt.cpu_affinity.mask == NULL) {
     perror("CPU_ALLOC");
     goto fail;
@@ -410,6 +425,24 @@ void parse_ionice(char *spec) {
   opt.ionice_prio = IOPRIO_PRIO_VALUE(n, data);
 }
 
+enum idmap parse_idmap(char *spec) {
+  const char *classes[] = {
+    [IDMAP_SELF] = "self",
+    [IDMAP_IDENTITY] = "identity",
+    [IDMAP_FULL] = "full",
+    NULL
+  };
+  enum idmap n;
+
+  for (n = 0; classes[n] && strcmp(spec, classes[n]); n++);
+  if (classes[n] == NULL) {
+    opt.error = true;
+    fprintf(stderr, "invalid idmap: %s\n", spec);
+    n = 0;
+  }
+  return n;
+}
+
 int sched_policy_from_name(const char *name) {
   if (!strcmp(name, "batch"))
     return SCHED_BATCH;
@@ -425,7 +458,6 @@ const struct option_info *find_option(int by_code,
                                       const char *by_name) {
   const struct option_info *optdef;
   const struct option_info *incompatible_option = NULL;
-  enum compat_level compat = opt.app->compat_level;
   bool is_short = (by_code && isascii(by_code)) ||
                   (by_name && *by_name && by_name[1] == '\0');
 
@@ -436,7 +468,7 @@ const struct option_info *find_option(int by_code,
 
     if ((by_code && by_code == (is_short ? optdef->short_name : (int) optdef->option)) ||
         (by_name && (is_short ? *by_name == optdef->short_name : optdef->long_name && !strcmp(by_name, optdef->long_name)))) {
-      if ((optdef->compat_level & compat) == 0) {
+      if (!is_compatible(optdef)) {
         /* Tentatively found an incompatible option but there may be
          * another one that is compatible later in the list. */
         incompatible_option = optdef;
@@ -589,15 +621,20 @@ static void read_options_file(const char *path) {
 
     optdef = find_option(0, option_name);
 
-    if (optdef && (optdef->compat_level & compat) == 0) {
+    if (optdef && !is_compatible(optdef)) {
       char short_name[2] = { optdef->short_name, 0};
       fprintf(stderr, "illegal option (%s) at this compat level\n",
               optdef->long_name ? optdef->long_name : short_name);
       opt.error = true;
       continue;
     } else if (!optdef) {
-      fprintf(stderr, "unknown option in config file: %s\n", option_name);
-      opt.error = true;
+      if (opt.optionality > 0) {
+        if (is_verbose())
+          fprintf(stderr, "ignoring unknown optional option: %s\n", option_name);
+      } else {
+        fprintf(stderr, "unknown option in config file: %s\n", option_name);
+        opt.error = true;
+      }
       continue;
     }
 
@@ -634,6 +671,12 @@ static void handle_option(enum compat_level *compat,
   case OPT_LEGACY:
     *compat = COMPAT_CHPST;
     break;
+  case OPT_OPT_START:
+    opt.optionality++;
+    break;
+  case OPT_OPT_END:
+    opt.optionality--;
+    break;
   case OPT_VERSION:
     opt.version = true;
     break;
@@ -702,6 +745,9 @@ static void handle_option(enum compat_level *compat,
   case OPT_NET_ADOPT:
     opt.net_adopt = optarg;
     break;
+  case OPT_IDMAP:
+    opt.idmap = parse_idmap(optarg);
+    break;
   case OPT_PRIVATE_RUN:
   case OPT_PRIVATE_TMP:
   case OPT_PROTECT_HOME:
@@ -709,6 +755,7 @@ static void handle_option(enum compat_level *compat,
   case OPT_RO_SYS:
   case OPT_RO_ETC:
   case OPT_PGRPHACK:
+  case OPT_FGHACK:
   case OPT_FORK_JOIN:
   case OPT_NEW_ROOT:
   case OPT_NO_NEW_PRIVS:
@@ -718,29 +765,21 @@ static void handle_option(enum compat_level *compat,
   case OPT_LOG_DIR:
   case OPT_LOGIN:
   case OPT_HARDLIMIT:
+  case OPT_DEV_NULL:
   case OPT_UGCLEARENV:
     /* Boolean options needing no further option processing */
     break;
 #ifndef NO_CAP
   case OPT_CAPBS_KEEP:
-    if (!parse_caps(&opt.cap_bounds, optarg))
-      opt.error = true;
-    opt.cap_bounds_op = CAP_OP_KEEP;
-    break;
   case OPT_CAPBS_DROP:
     if (!parse_caps(&opt.cap_bounds, optarg))
       opt.error = true;
-    opt.cap_bounds_op = CAP_OP_DROP;
     break;
   case OPT_CAPS_KEEP:
-    if (!parse_caps(&opt.caps, optarg))
-      opt.error = true;
-    opt.caps_op = CAP_OP_KEEP;
-    break;
   case OPT_CAPS_DROP:
     if (!parse_caps(&opt.caps, optarg))
       opt.error = true;
-    opt.caps_op = CAP_OP_DROP;
+    opt.caps_op = optdef->option == OPT_CAPS_KEEP ? CAP_OP_KEEP : CAP_OP_DROP;
     break;
 #endif
   case OPT_CPU_SCHED:
@@ -883,7 +922,10 @@ static void handle_option(enum compat_level *compat,
   case OPT_CLOSE_STDIN:
   case OPT_CLOSE_STDOUT:
   case OPT_CLOSE_STDERR:
-    opt.close_fds |= 1 << (optdef->option - OPT_CLOSE_STDIN + STDIN_FILENO);
+    if (set(OPT_DEV_NULL))
+      opt.dev_null_fds |= 1 << (optdef->option - OPT_CLOSE_STDIN + STDIN_FILENO);
+    else
+      opt.close_fds |= 1 << (optdef->option - OPT_CLOSE_STDIN + STDIN_FILENO);
     break;
   /* Avoid default case to get useful compiler warnings instead.
   default:
@@ -907,7 +949,7 @@ int options_parse(int argc, char *argv[]) {
 
     optdef = find_option(c, NULL);
 
-    if (optdef && (optdef->compat_level & compat) == 0) {
+    if (optdef && !is_compatible(optdef)) {
       char short_name[2] = { optdef->short_name, 0};
       fprintf(stderr, "illegal option (%s) at this compat level\n",
               optdef->long_name ? optdef->long_name : short_name);
