@@ -14,7 +14,9 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#ifndef NO_CAP
 #include <sys/capability.h>
+#endif
 #include <sys/resource.h>
 #include <linux/ioprio.h>
 
@@ -32,6 +34,7 @@ enum compat_level {
   COMPAT_SETLOCK           = 00200,
   COMPAT_APPLYUIDGID       = 00400,
   COMPAT_SETUIDGID_FROMENV = 01000,
+  COMPAT_FGHACK            = 02000,
 };
 
 
@@ -53,7 +56,7 @@ static const/*expr*/ enum compat_level C_L = COMPAT_SETLOCK;
 
 static const/*expr*/ enum compat_level C_A = COMPAT_APPLYUIDGID;
 
-static const/*expr*/ enum compat_level C_ALL = 01777;
+static const/*expr*/ enum compat_level C_ALL = 03777;
 
 enum opt /* C23: :int */ {
   OPT_BASE = 0x1000,
@@ -93,7 +96,10 @@ enum opt /* C23: :int */ {
   OPT_VERBOSE,
   OPT_VERSION,
   OPT_PGRPHACK,
+  OPT_FGHACK,
   OPT_LEGACY,
+  OPT_OPT_START,
+  OPT_OPT_END,
   OPT_HELP,
   OPT_FILE,
   OPT_MOUNT_NS,
@@ -134,6 +140,8 @@ enum opt /* C23: :int */ {
   OPT_GIDSNUM,
   OPT_UGFROMENV,
   OPT_UGCLEARENV,
+  OPT_IDMAP,
+  OPT_DEV_NULL,
 
   /* Keep at end */
   OPT_EXIT,
@@ -176,6 +184,12 @@ enum cap_op {
   CAP_OP_DROP,
 };
 
+enum idmap {
+  IDMAP_SELF = 0,
+  IDMAP_IDENTITY,
+  IDMAP_FULL,
+};
+
 struct options_file {
   struct options_file *next;
   char content[];
@@ -198,6 +212,7 @@ struct options {
   bool exit;
   int retcode;
   int verbosity;
+  int optionality;
 
   /* The interesting (x)chpsty stuff here */
   char *argv0;
@@ -233,6 +248,7 @@ struct options {
   struct limit rlimit_sigpending;
   struct limit rlimit_locks;
   unsigned close_fds;
+  unsigned dev_null_fds;
 #ifndef NO_CAP
   enum cap_op cap_bounds_op;
   cap_bits_t cap_bounds;
@@ -241,6 +257,7 @@ struct options {
 #endif
   unsigned int umask;
   long oom_adjust;
+  enum idmap idmap;
 
   struct {
     cpu_set_t *mask;
@@ -250,6 +267,10 @@ struct options {
 
 extern struct options opt;
 
+static inline bool is_compatible(const struct option_info *option) {
+  return option->compat_level & opt.app->compat_level;
+}
+
 static inline bool is_verbose(void) {
   return opt.verbosity >= LOG_LEVEL_VERBOSE;
 }
