@@ -16,18 +16,19 @@
 #include <sched.h>
 #include <signal.h>
 #include <stdarg.h>
+#include <strings.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
-#include <linux/prctl.h>
+#include <libgen.h>
 #include <linux/ioprio.h>
 #include <sys/file.h>
 #include <sys/dir.h>
+#include <sys/eventfd.h>
 #include <sys/mount.h>
 #include <sys/stat.h>
-#include <sys/pidfd.h>
 #include <sys/prctl.h>
 #include <sys/signalfd.h>
 #include <sys/time.h>
@@ -42,6 +43,7 @@
 #include "rootfs.h"
 #include "mount.h"
 #include "precreate.h"
+#include "dirs.h"
 
 static const char *version_str = STRINGIFY(PROG_VERSION);
 #ifdef PROG_DEFAULT
@@ -49,7 +51,6 @@ static const char *default_app = STRINGIFY(PROG_DEFAULT);
 #else
 static const char *default_app = "xchpst";
 #endif
-#define PROG "xchpst"
 
 const struct app apps[] = {
   { COMPAT_CHPST,       "chpst",     .long_opts = false },
@@ -57,6 +58,7 @@ const struct app apps[] = {
   { COMPAT_SOFTLIMIT,   "softlimit", .long_opts = false },
   { COMPAT_ENVDIR,      "envdir",    false, 1, { OPT_ENVDIR } },
   { COMPAT_PGRPHACK,    "pgrphack",  false, 1, { OPT_PGRPHACK } },
+  { COMPAT_FGHACK,      "fghack",    false, 1, { OPT_FGHACK } },
   { COMPAT_SETUIDGID,   "setuidgid", false, 1, { OPT_SETUIDGID } },
   { COMPAT_ENVUIDGID,   "envuidgid", false, 1, { OPT_ENVUIDGID } },
   { COMPAT_SETLOCK,     "setlock",   false, 1, { OPT_LOCK_WAIT } },
@@ -71,12 +73,12 @@ static const char *suite_prefixes[] = {
   NULL
 };
 
-struct runtime runtime = { 0 };
-
-static const char *std_run_dir = "/run/" PROG;
-static const char *fallback_run_dir = "/tmp/run-" PROG;
-static int run_dir_fd = -1;
-char *run_dir;
+static int proc_self_fd = -1;
+static struct {
+  int needed;
+  int sync_fd;
+  pid_t pid;
+} nshelper = { 0 };
 
 static struct {
   char uid[16];
@@ -100,50 +102,33 @@ static void usage(FILE *out) {
   options_print(out);
 }
 
-int ensure_dir(int dirfd, const char *path, int *fd, mode_t mode) {
-  for (int creating = 0; *fd == -1 && creating <= 1; creating++) {
-    *fd = openat(dirfd, path, O_DIRECTORY | O_CLOEXEC);
-    if (*fd == -1 && !creating && (errno == ENOENT))
-      mkdirat(dirfd, path, mode);
-  }
-  if (*fd == -1 && errno != EEXIST)
-    return -1;
-  return 0;
+static int open_proc_self(void) {
+  if (proc_self_fd == -1)
+    proc_self_fd = open("/proc/self", O_RDONLY | O_DIRECTORY | O_CLOEXEC);
+  else if (proc_self_fd == -1)
+    perror("opening /proc/self");
+  return proc_self_fd;
 }
 
-int get_run_dir(void) {
-  int rc = ensure_dir(-1, std_run_dir, &run_dir_fd, 0700);
-  if (rc == -1) {
-    char *xdg_run_dir = getenv("XDG_RUNTIME_DIR");
-    if (xdg_run_dir) {
-      char *path;
-      rc = asprintf(&path, "%s/%s", xdg_run_dir, PROG);
-      if (rc != -1) {
-        rc = ensure_dir(-1, path, &run_dir_fd, 0700);
-        if (rc != -1) {
-          run_dir = path;
-        } else {
-          free(path);
-        }
-      }
-    }
-  } else {
-    run_dir = strdup(std_run_dir);
+static void close_proc_self(void) {
+  if (proc_self_fd != -1) {
+    if (close(proc_self_fd) == -1)
+      perror("closing /proc/self");
+    proc_self_fd = -1;
   }
-  if (rc == -1 &&
-      ensure_dir(-1, fallback_run_dir, &run_dir_fd, 0700) != -1)
-    run_dir = strdup(fallback_run_dir);
-
-  return run_dir_fd;
 }
 
-static int write_once(const char *file, const char *fmt, ...) {
-  int fd = open(file, O_WRONLY);
+static int write_proc_self_once(const char *file, const char *fmt, ...) {
+  int fd;
   char *text;
   ssize_t len;
   int rc = 0;
   va_list args;
 
+  if (open_proc_self() == -1)
+    goto fail0;
+
+  fd = openat(proc_self_fd, file, O_WRONLY);
   if ((rc = fd == -1 ? 1 : 0))
     goto fail0;
 
@@ -168,6 +153,45 @@ fail0:
   return rc;
 }
 
+void set_idmap(const char *path, enum idmap map, long id) {
+  switch (map) {
+  case IDMAP_SELF:
+    write_proc_self_once(path, "%lu %lu %lu\n", 0, id, 1);
+    break;
+  case IDMAP_IDENTITY:
+    write_proc_self_once(path, "%lu %lu %lu\n", 0, 0, 0x00010000UL);
+    break;
+  case IDMAP_FULL:
+    write_proc_self_once(path, "%lu %lu %lu\n", 0, 0, 0xFFFFFFFFUL);
+    break;
+  }
+}
+
+int ns_helper(gid_t gid, uid_t uid) {
+  if (nshelper.pid > 0) {
+    const uint64_t e = 1;
+    int rc, status;
+    do rc = write(nshelper.sync_fd, &e, sizeof e);
+    while (rc == -1 && errno == EINTR);
+    if (rc == -1) {
+      perror("ns_helper: write()");
+      return 1;
+    }
+    do rc = waitpid(nshelper.pid, &status, 0);
+    while (rc == -1 && errno == EINTR);
+    if (rc == -1) {
+      perror("ns_helper: waitpid()");
+      return 1;
+    }
+    if (WIFEXITED(status) && WEXITSTATUS(status) != EXIT_SUCCESS)
+      return 1;
+  } else {
+    set_idmap("gid_map", opt.idmap, gid);
+    set_idmap("uid_map", opt.idmap, uid);
+  }
+  return 0;
+}
+
 void set_rlimit(int resource, struct limit *option) {
   struct rlimit prev;
 
@@ -310,6 +334,13 @@ int main(int argc, char *argv[]) {
   if (is_verbose())
     fprintf(stderr, "invoked as %s(%s)\n", opt.app->name, program_invocation_short_name);
 
+  if (!set(OPT_FORK_JOIN) &&
+      set(OPT_FGHACK)) {
+    if (is_verbose())
+      fprintf(stderr, "also going to do fork-join since fghack requested\n");
+    enable(OPT_FORK_JOIN);
+  }
+
   if (!set(OPT_FORK_JOIN) &&
       (opt.new_ns & CLONE_NEWPID)) {
     if (is_verbose())
@@ -320,7 +351,7 @@ int main(int argc, char *argv[]) {
   if (!(opt.new_ns & CLONE_NEWNS) &&
       (set(OPT_NET_NS) || set(OPT_NET_ADOPT) || set(OPT_PRIVATE_RUN) || set(OPT_PRIVATE_TMP) ||
        set(OPT_RO_SYS) || set(OPT_RO_HOME) || set(OPT_RO_ETC) ||
-       set(OPT_NEW_ROOT) || set(OPT_PID_NS))) {
+       set(OPT_NEW_ROOT) || opt.new_ns & CLONE_NEWPID)) {
     if (is_verbose())
       fprintf(stderr, "also creating mount namespace implicitly due to other options\n");
     opt.new_ns |= CLONE_NEWNS;
@@ -346,7 +377,8 @@ int main(int argc, char *argv[]) {
 
   if (opt.error) {
     const struct option_info *help_option = find_option(OPT_HELP, NULL);
-    if (help_option && opt.app->long_opts && help_option->long_name)
+    if (help_option && opt.app->long_opts && help_option->long_name &&
+        is_compatible(help_option))
       fprintf(stderr, "%s: error in options. Run %s --%s for usage\n",
               program_invocation_short_name,
               program_invocation_short_name,
@@ -371,7 +403,7 @@ int main(int argc, char *argv[]) {
     umask(opt.umask);
 
   if (set(OPT_OOM))
-    write_once("/proc/self/oom_score_adj", "%ld", opt.oom_adjust);
+    write_proc_self_once("oom_score_adj", "%ld", opt.oom_adjust);
 
   if (opt.lock_file) {
     if (opt.lock_nowait_override)
@@ -580,13 +612,37 @@ int main(int argc, char *argv[]) {
       !drop_user(&opt.users_groups))
       goto finish;
 
+  if (opt.idmap != IDMAP_SELF)
+    nshelper.needed++;
+
+  if (nshelper.needed) {
+    uint64_t e;
+    if (open_proc_self() == -1 ||
+        (nshelper.sync_fd = eventfd(0, 0)) == -1 ||
+        (nshelper.pid = fork()) == -1) {
+      perror("ns_helper: open(\"/proc/self\"); eventfd(); fork()");
+      goto finish;
+    }
+    if (nshelper.pid == 0) {
+      while ((rc = read(nshelper.sync_fd, &e, sizeof e)) == -1 &&
+             (errno == EAGAIN || errno == EINTR));
+      if (rc != sizeof e || e != 1) {
+        perror("ns_helper: read() and sync");
+        goto finish;
+      }
+      exit(ns_helper(gid, uid));
+    }
+    /* Any state (variables) relied upon by the namespace helper must not
+     * change between here and the sync. */
+  }
+
   if (opt.new_ns) {
     rc = unshare(opt.new_ns);
     if (rc == -1) {
       perror(NAME_STR ": unshare()");
       goto finish;
     }
-    if (opt.verbosity > 0) fprintf(stderr, "created 0xb%b namespaces\n", opt.new_ns);
+    if (is_debug()) fprintf(stderr, "created 0x%x namespaces\n", opt.new_ns);
 
     if (opt.new_ns & CLONE_NEWNS) {
       rc = mount(NULL, "/", NULL,
@@ -599,11 +655,11 @@ int main(int argc, char *argv[]) {
   if (opt.new_ns & CLONE_NEWUSER) {
     rc = prctl(PR_SET_DUMPABLE, 1);
     setgroups(0, NULL);
-    write_once("/proc/self/setgroups", "%s", "deny\n");
-    write_once("/proc/self/gid_map", "%u %u %u\n", 0, gid, 1);
-    write_once("/proc/self/uid_map", "%u %u %u\n", 0, uid, 1);
-    if (setresgid(0, 0, 0) != 0 ||
-        setresuid(0, 0, 0) != 0)
+    write_proc_self_once("setgroups", "%s", "deny\n");
+    ns_helper(gid, uid);
+    if (opt.idmap == IDMAP_SELF &&
+        (setresgid(0, 0, 0) != 0 ||
+         setresuid(0, 0, 0) != 0))
       fprintf(stderr, "warning: error becoming root in user namespace, %s\n", strerror(errno));
   }
 
@@ -639,17 +695,13 @@ adopt_fail0:
     special_mount("/sys", "sysfs", "sysfs", NULL);
 
   if (set(OPT_FORK_JOIN)) {
-    /* Save old signal mask for re-use by child and block all signals
-     * in the parent so we can get them delivered by signalfd. */
-    sigfillset(&newmask);
-    sigdelset(&newmask, SIGCHLD);
-    sigdelset(&newmask, SIGBUS);
-    sigdelset(&newmask, SIGFPE);
-    sigdelset(&newmask, SIGILL);
-    sigdelset(&newmask, SIGSEGV);
-    rc = sigprocmask(SIG_SETMASK, &newmask, &oldmask);
-    if (rc == -1) {
-      perror("setting up mask for signalfds");
+    bool subreaper = set(OPT_FGHACK) && !(opt.new_ns & CLONE_NEWPID);
+
+    if (!sig_proxy_mask(&newmask, &oldmask))
+      goto finish;
+
+    if (subreaper && prctl(PR_SET_CHILD_SUBREAPER, 1) == -1) {
+      perror("becoming subpreaper for fghack");
       goto finish;
     }
 
@@ -760,8 +812,28 @@ adopt_fail0:
       goto finish;
 #endif
 
+  if (set(OPT_FGHACK) &&
+      opt.new_ns & CLONE_NEWPID &&
+      !pico_init(open_proc_self()))
+    goto finish;
+
+  if (opt.dev_null_fds) {
+    int devnull = open("/dev/null", 0);
+    if (devnull == -1) {
+      perror("open /dev/null");
+      goto finish;
+    }
+    for (unsigned int fds = opt.dev_null_fds; fds; fds &= ~(1 << fd))
+      if (dup2(devnull, fd = ffs(fds) - 1) == -1) {
+        perror("dup2");
+        close(devnull);
+        goto finish;
+      }
+    close(devnull);
+  }
+
   for (unsigned int close_fds = opt.close_fds; close_fds; close_fds &= ~(1 << fd))
-    close(fd = /*stdc_trailing_zeros*/ __builtin_ctz(close_fds));
+    close(fd = ffs(close_fds) - 1);
 
   if (set(OPT_NO_NEW_PRIVS) && prctl(PR_SET_NO_NEW_PRIVS, 1L, 0L, 0L, 0L) == -1)
     perror("could not honour --no-new-privs");
@@ -779,7 +851,9 @@ adopt_fail0:
 
 join:
   if (set(OPT_FORK_JOIN) && child != 0)
-    join(child, &newmask, &oldmask, &ret);
+    join(child, &newmask, &oldmask, &ret,
+         set(OPT_FGHACK) && !(opt.new_ns & CLONE_NEWPID),
+         open_proc_self(), "fork-join");
 
 finish:
   /* Actions here should be
@@ -788,17 +862,16 @@ finish:
      3) not be necessary when --fork-join is not used.
    */
 
+  close_proc_self();
   free_rootfs_data();
   pivot_tidy(&pivot);
-  if (run_dir_fd != -1)
-    close(run_dir_fd);
+  close_run_dir();
 
   if (lock_fd != -1)
     close(lock_fd);
 
   free(pwd);
   free(sub_argv);
-  free(run_dir);
   free(extra_env.gidlist);
 
 finish0:
